#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <assert.h>

#include <libsi/file.h>

typedef struct {
    uint64_t start;
    uint64_t end;
    char perm[5];
    uint64_t offset;
    int major;
    int minor;
    int inode;
    char* path;
} proc_maps_ent;

typedef struct {
    char* all;
    proc_maps_ent* ents;
    int ent_size;
} proc_maps;

proc_maps* proc_maps_read(pid_t pid) {
    proc_maps* pm = (proc_maps*)calloc(1, sizeof(proc_maps));
    char buf[256];
    char* p;
    int i;

    sprintf(buf, "/proc/%d/maps", pid);
    pm->all = file_read(buf);

    pm->ent_size = 0;
    for (p = pm->all; *p; p++) {
        if (*p == '\n') pm->ent_size++;
    }
    pm->ents = (proc_maps_ent*)calloc(pm->ent_size, sizeof(proc_maps_ent));

    p = pm->all;
    for (i = 0; i < pm->ent_size; i++) {
        proc_maps_ent* e = &pm->ents[i];
        int consumed;
        if (sscanf(p, "%llx-%llx %s %llx %x:%x %d %n",
                   &e->start, &e->end, e->perm, &e->offset,
                   &e->major, &e->minor, &e->inode, &consumed) < 0) {
            pm->ent_size = i;
            break;
        }
        p += consumed;
        for (consumed = 0; p[consumed] && p[consumed] != '\n'; consumed++) {}
        e->path = (char*)malloc(consumed+1);
        memcpy(e->path, p, consumed);
        p += consumed;
    }

    return pm;
}

proc_maps_ent* proc_maps_ent_by_addr(proc_maps* pm, uint64_t addr) {
    int i;
    assert(pm);
    for (i = 0; i < pm->ent_size; i++) {
        proc_maps_ent* e = &pm->ents[i];
        if (e->start <= addr && addr < e->end) {
            return e;
        }
    }
    return NULL;
}

void proc_maps_free(proc_maps* pm) {
    int i;
    assert(pm);
    for (i = 0; i < pm->ent_size; i++) {
        free(pm->ents[i].path);
    }
    free(pm->all);
    free(pm);
}
