#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <sys/user.h>
#include <sys/resource.h>

#include <libsi/proc.h>

int main(int argc, char* argv[]) {
    if (argc == 1) {
        printf("Usage: %s <binary>\n", argv[0]);
        return 1;
    }

    int pid;
    if ((pid = fork())) {
        int steps = 0;
        void* prip = NULL;
        struct user_regs_struct regs;
        proc_maps* pm = NULL;

        while (1) {
            int status;
            wait(&status);

            if (WIFEXITED(status)) {
                fprintf(stderr, "%s\n", pm->all);
                proc_maps_free(pm);

                struct rusage ru;
                getrusage(RUSAGE_CHILDREN, &ru);
                fprintf(stderr, "user: %d.%06d, sys: %d.%06d\n",
                        (int)ru.ru_utime.tv_sec, (int)ru.ru_utime.tv_usec,
                        (int)ru.ru_stime.tv_sec, (int)ru.ru_stime.tv_usec);

                fprintf(stderr, "%d steps.\n", steps);
                return 0;
            }
            if (WIFSIGNALED(status)) {
                fprintf(stderr, "killed by signal\n");
                return 1;
            }

            steps++;
            if (ptrace(PTRACE_GETREGS, pid, 0, &regs) < 0) {
                perror("PTRACE_GETREGS");
                return 1;
            }
            void* rip = (void*)regs.rip;

            proc_maps_ent* ent = NULL;
            if (!pm || !(ent = proc_maps_ent_by_addr(pm, (uint64_t)rip))) {
                if (pm) {
                    proc_maps_free(pm);
                }
                pm = proc_maps_read(pid);
                ent = proc_maps_ent_by_addr(pm, (uint64_t)rip);
            }

            if (ptrace(PTRACE_SINGLESTEP, pid, 0, 0) < 0) {
                perror("PTRACE_SINGLESTEP");
                return 1;
            }

            fprintf(stderr, "%p => %p (%s @%llx)\n",
                    prip, rip, ent->path, rip - ent->start);
            prip = rip;
        }
    }
    else {
        ptrace(PTRACE_TRACEME, 0, 0, 0);
        execl(argv[1], argv[1], NULL);
    }
}
