def parse_objdump(file)
  obj = {}
  func = nil
  offset = nil
  IO.popen("objdump -d -F #{file}", 'r') do |pipe|
    pipe.each do |line|
      if /^([0-9a-f]+) <(.*?)> \(.*?: 0x([0-9a-f]+)\):/ =~ line
        func = $2
        offset = $1.hex - $3.hex
      elsif /^\s*([0-9a-f]+):\s*(([0-9a-f]{2} )+)\s*(.*)/ =~ line
        obj[$1.hex - offset] = [func, $2.split.size, $4]
      end
    end
  end
  obj
end

class Stat
  def initialize
    @steps = 0
    @bytes = 0
    @ops = Hash.new{0}
    @addr_to_size = {}
  end

  def add(size, op, abs)
    @steps += 1
    @bytes += size
    @ops[op.split[0]] += 1
    @addr_to_size[abs] = size
  end

  def to_s
    area = @addr_to_size.inject(0){|r,a|r+a[1]}
    "executed #@steps steps, #@bytes bytes in #{area} bytes"
  end

  def opstat
    @ops.sort_by{|op, cnt|-cnt}.map{|op, cnt|"#{op} #{cnt}"} * "\n"
  end
end

f = File.open('trace')

objs = {}
bytes = 0
cur_func = nil
cur_stat = Stat.new
cur_file = nil
file_stat = {}
total_stat = Stat.new

f.each do |l|
  if / => 0x([0-9a-f]+) \((.*?) @([0-9a-f]+)\)/ !~ l
    next
  end

  abs = $1.hex
  file = $2
  rel = $3.hex

  if !objs[file]
    objs[file] = parse_objdump(file)
  end

  obj = objs[file]
  func, size, op = obj[rel]
  if cur_func != nil && cur_func != func
    puts "Stat of #{cur_func} (#{cur_file}): #{cur_stat.to_s}"
    cur_stat = Stat.new
  end
  if !file_stat[file]
    file_stat[file] = Stat.new
  end
  file_stat[file].add(size, op, abs)
  cur_stat.add(size, op, abs)
  total_stat.add(size, op, abs)
  cur_file = file
  cur_func = func
  bytes += size

  puts "#{l.chomp} #{func}"
  puts "\t#{op}"
end

puts "Stat of #{cur_func} (#{cur_file}): #{cur_stat.to_s}"

f.close

file_stat.each do |f, st|
  puts "Stat of #{f}: #{st.to_s}"
end

puts total_stat.to_s
puts total_stat.opstat

